-- global constructor
function createObj(template, init) 
	local new = init or {}
	for k,v in template do 
		if not new[k] then new[k]=v end
	end
	return new
end

STATUS_ALIVE=0
STATUS_DEAD=1
STATUS_FLEEING=2
STATUS_IMPROVISE=3

function rotate180(d)
	return mod(d+2,4);
end

function rotate90(d)
	return mod(d+1,4);
end

function manhattan_distance(x1,y1,x2,y2)
	local dx=x1-x2
	local dy=y1-y2
	if dx<0 then dx=-dx end
	if dy<0 then dy=-dy end
	return dx+dy
end

function dir2vect(d)
	local dx=0
	local dy=0
	if d==O_EAST then dx=1
	elseif d==O_WEST then dx=-1
	elseif d==O_NORTH then dy=-1
	elseif d==O_SOUTH then dy=1
	end
	return dx,dy
end

function vect2dir(dx,dy)
	if dx>0 then 
		return O_EAST
	elseif dx<0 then 
		return O_WEST 
	end
	
	if dy>0 then 
		return O_SOUTH
	elseif dy<0 then 
		return O_NORTH 
	end
	return O_EAST -- ...
end

function is_safe_place(tx, ty)
	local o=nil
--	write('is_safe_place('..tx..','..ty..')\n')
	for dir=0,3 do
		o=Field:get_first_object(tx,ty,dir)
		if o ~= nil then
			if o:type() == OBJ_BOMB then
				local b=o:to_bomb()
				local r=b:get_radius()
				local bx=Field:x_to_tile(b:get_x())
				local by=Field:y_to_tile(b:get_y())
				local d=manhattan_distance(tx,ty,bx,by)
--				write('found bomb r='..r..' d='..d..'\n')
				if r >= d then 
					return nil 
				end
			end
		end
	end
	return 1
end

----------------------------------------------------
-- This is a bot that makes circles....
---------------------------------------------------- 

CircleDude = {
	so = nil,
	count = 0
}

function CircleDude.new(script_obj)
	local self=createObj(script_obj)
	self.so=script_obj
	return self
end

function CircleDude:think()
	if not bot:is_moving() then
		if count == nil then count = 0 end
		if count == 0 then
			bot:horizontal_slide(self.orgspeed, 500);
		elseif count==1 then
			bot:vertical_slide(self.orgspeed, 200);
		elseif count==2 then
			bot:horizontal_slide(self.orgspeed, 0);
		elseif count==3 then
			bot:vertical_slide(self.orgspeed, 40);
		end
		count=count+1
		if count>3 then count=0 end
	end
end

----------------------------------------------------
---- Dumbo avoids bombs with a movemap
---------------------------------------------------- 

Dumbo = {
	so = nil, 			-- scriptobject
	status=STATUS_ALIVE, -- status
	orgspeed=250/1000,  -- travelling speed in px/s
	dc=0,				-- death counter
	orgdc=1000,			-- initial value for death counter
	lasttime=0,			-- time of last direction change
	deaths=0,			-- # times the Dumbo died
	maxdeaths=3 		-- # times the Dumbo can die before its removed 
						-- from the field
}

function Dumbo.new(script_obj)
	local self=createObj(Dumbo)
	self.so=script_obj
	return self
end


movemap = {}
movemap[0] = { 3, 1, 2}
movemap[1] = { 2, 3, 0}
movemap[2] = { 3, 1, 2}
movemap[3] = { 2, 3, 1}

function Dumbo:choosedir()
	local bot=self.so
	local dx,dy=dir2vect(bot:get_direction())
	dx=dx*20
	dy=dy*20
	local tx=Field:x_to_tile(bot:get_x()+dx)
	local ty=Field:y_to_tile(bot:get_y()+dy)

	local dir=bot:get_direction()
	local o=Field:get_first_object(tx,ty,dir)
	if o ~= nil and o:type() == OBJ_FLAME then
		bot:slide(rotate180(dir),self.orgspeed,40)
	end

	for td=0,3 do
--		local td=random(0,3)
		o=Field:get_first_object(tx, ty, mod(dir+td,4))
		if o~=nil then
			if o:type() == OBJ_BOMB then
				local b = o:to_bomb()
				local fx = Field:x_to_tile(b:get_x())
				local fy = Field:y_to_tile(b:get_y())
				if b:get_radius() >= manhattan_distance(tx, ty, fx, fy) then
					--for _,d in movemap[td] do
						d = movemap[td][random(1,3)]
						bot:slide(mod(dir+d,4),self.orgspeed,random(2,4)*40)
						if bot:is_sliding() then
							return
						end
					--end
				end
			end
		end
	end
end

function Dumbo:think()
	local bot = self.so
	local s = self.orgspeed
	local x = bot:get_x()
	local y = bot:get_y()
	local was_sliding = bot:is_sliding()

	if self.status == STATUS_ALIVE then
		if bot:is_sliding() then
			self:choosedir()
		end
		
		if not bot:is_sliding() and random() > .7 then
			local dir=random(0,3)
			local distance=random(2,4)*40
			bot:slide(dir,s,distance)
		end

		-- remember time of last direction change
		if bot:is_sliding() and not was_sliding then
				self.lasttime=Game:get_ticks()
				bot:setAnimation(0)
		end

		---------------------------------------------------- 
		-- do stuff with objects
		--local d=bot:get_direction()
		--if d==O_NORTH then dy=-20
		--local o=Field:get_object_at_xy(x+dx, y+dy)
		--if o ~= nil then self:handleObject(o) end
		--local ot = o:type()
		--if ot==OBJ_BOMB or ot==OBJ_BONUS then...

	else -- self.status = STATUS_DEAD
		self.dc=self.dc-Game:get_elapsed_time()
		if self.dc < 0 then 
			if self.deaths>=self.maxdeaths then
				-- die for good
				self.so:set_remove_flag()
			else
				-- live again
				self.status = 0
				bot:setAnimation(0)
			end
		end
	end
end

function Dumbo:die()
	if self.status == STATUS_DEAD then
		-- don't die when already dead
		return
	else
		self.deaths=self.deaths+1
		self.dc = self.orgdc
		self.status = 1
		self.so:stop()
		self.so:setAnimation(1)
	end
end


-------------------------------------------
-- Ganjawarrior makes a map of dangerous places in his head, then tries 
-- to find a close safe place and move to it
-------------------------------------------

GanjaWarrior = {
	so = nil, 			-- scriptobject
	status=STATUS_ALIVE, -- status
	orgspeed=350/1000,  -- travelling speed in px/s
	dc=0,				-- death counter
	orgdc=1000,			-- initial value for death counter
	lasttime=0,			-- time of last direction change
	deaths=0,			-- # times the Dumbo died
	maxdeaths=3, 		-- # times the Dumbo can die before its removed 
						-- from the field
	safe_x=-1, 			-- where to go if status=FLEEING
	safe_y=-1
}

function GanjaWarrior.new(script_obj)
	local self=createObj(GanjaWarrior)
	self.so=script_obj
	return self
end

function GanjaWarrior:here_is_safe()
	local bot=self.so
	local tx=Field:x_to_tile(bot:get_x())
	local ty=Field:y_to_tile(bot:get_y())
	return is_safe_place(tx,ty)
end

function GanjaWarrior:find_safe_place()
	local bot=self.so
	local x=bot:get_x()
	local y=bot:get_y()
	local tx=Field:x_to_tile(x)
	local ty=Field:y_to_tile(y)

	-- mark dangerous places
	local o
	local f={}

	for j=ty-5,ty+5 do
		f[j]={}
		for i=tx-5,tx+5 do
			f[j][i]=0
		end
	end
	
	-- 0 = free
	-- 1 = wall
	-- 2 = flame
	-- 3 = bomb
	-- 4 = future flame
	-- 5 = other object
	-- 6 = player
	for j=ty-5,ty+5 do
		for i=tx-5,tx+5 do
			if Field:get_tile_at(i,j) ~= TILE_TYPE_EMPTY then
				f[j][i]=1
			end
			
			o=Field:get_object_at(i,j)
			if o~=nil then
				if o:type() == OBJ_FLAME then
					f[j][i]=2
				elseif o:type() == OBJ_BOMB then
					f[j][i]=3
					local b=o:to_bomb()
					local r=b:get_radius()
					local im = i-r; if im<tx-5 then im=tx-5 end
					local iM = i+r; if iM>tx+5 then iM=tx+5 end
					for i1=im,iM do f[j][i1]=4 end
					local jm = j-r; if jm<ty-5 then jm=ty-5 end
					local jM = j+r; if jM>ty+5 then jM=ty+5 end
					for j1=jm,jM do f[j1][i]=4 end
				else
					f[j][i]=5
				end
			end
		end
	end

	f[ty][tx]=6
	
	for j=ty-5,ty+5 do
		for i=tx-5,tx+5 do
			write(f[j][i])
		end
		write('\n')
	end
	
	-- using a spiral algorithm to find the closest safe places first
	local dir=1  
	local nx=tx      -- tile we check
	local ny=ty-1
	local c=0       -- 
	local C=0 		--
	local n=0 		-- 
	local N=24 
	local S=0
	local dx,dy
	dx=0
	dy=0
	
	write('starting spiral from tx='..tx..' ty='..ty..'\n')
	
	while n<N do
		for s=0,S do
			for c=0,C do
--				write('dx='..dx..' dy='..dy..' nx='..nx..' ny='..ny..' n='..n..'\n')
				write('checking '..nx..','..ny..'\n')
				if f[ny]~=nil and f[ny][nx] == 0 then
					write('found first safe heaven at nx='..nx..' ny='..ny..'\n')
					self.safe_x = Field:tile_to_x(nx)
					self.safe_y = Field:tile_to_y(ny)
					self.lasttime = Game:get_ticks()
					return
				end
				dx,dy=dir2vect(dir)
				nx=nx+dx
				ny=ny+dy
				n=n+1
			end
			dir=mod(dir+1,4)
			if n>N then break end
		end
		C=C+1
		S=1
	end

	write('holy crap ! i\'m in deep shit now !\n')
end
	

function GanjaWarrior:think()
	local bot=self.so

	if self.status == STATUS_IMPROVISE then
		write('bot is improvising.\n')
		if not bot:is_moving() then
			self.status = STATUS_FLEEING
		end
	end
	
	if self.status == STATUS_ALIVE then
		if not self:here_is_safe() then
			write('uh oh.. here is not safe !\n')
			self.status = STATUS_FLEEING
			bot:stop()
			self:find_safe_place()
		else
--			write('this place looks safe..\n')
		end
	end

	if self.status == STATUS_FLEEING then
--		write('bot is fleeing to '..self.safe_x..','..self.safe_y..'\n')
		
		if self.safe_x < 0 or self.safe_y < 0 then 
			write('<0\n')
			self.status = STATUS_ALIVE 
			return
		end

		if Game:get_ticks() - self.lasttime > 1000 then
			write('reset target after 1 second\n')
			bot:stop()
			self:find_safe_place()
		end

		if Field:x_to_tile(bot:get_x()) == Field:x_to_tile(self.safe_x) and
		   Field:y_to_tile(bot:get_y()) == Field:y_to_tile(self.safe_y)
		then
			write('bot has reached target\n')
			bot:stop()
			bot:set_x(Field:center_x(bot:get_x()))
			bot:set_y(Field:center_x(bot:get_y()))
			self.status = STATUS_ALIVE
		end
		
		if not is_safe_place(Field:x_to_tile(self.safe_x), Field:y_to_tile(self.safe_y)) then
			write('safe place is not safe anymore, recalculating!\n')
			bot:stop()
			self:find_safe_place()
			self.status = STATUS_FLEEING
		end
		
		if self.status == STATUS_FLEEING and not bot:is_moving() then
			local dx=self.safe_x - bot:get_x()
			local dy=self.safe_y - bot:get_y()
			if abs(dx)>=abs(dy) then
				dir=vect2dir(dx,0)
				bot:slide(dir,self.orgspeed,abs(dx))
				if not bot:is_moving() then
					dir=vect2dir(0,dy)
					bot:slide(dir,self.orgspeed,abs(dy))
				end
			else
				dir=vect2dir(0,dy)
				bot:slide(dir,self.orgspeed,abs(dy))
				if not bot:is_moving() then
					dir=vect2dir(dx,0)
					bot:slide(dir,self.orgspeed,abs(dx))
				end
			end

			if not bot:is_moving() then
				write('plan blocked, improvising\n')
				for i=0,3 do
					bot:slide(i,self.orgspeed,40)
					if bot:is_moving() then break end
				end
				self.status = STATUS_IMPROVISE
			end

			if not bot:is_moving() then
				write('uh oh ! bot is going to die :)\n')
			end
--			write('fleeing to '..dir)
		end
	end

	if self.status == STATUS_DEAD then -- self.status = STATUS_DEAD
		self.dc=self.dc-Game:get_elapsed_time()
		if self.dc < 0 then 
			if self.deaths>=self.maxdeaths then
				-- die for good
				self.so:set_remove_flag()
			else
				-- live again
				self.status = 0
				bot:setAnimation(0)
			end
		end
	end
end

function GanjaWarrior:die()
	if self.status == STATUS_DEAD then
		-- don't die when already dead
		return
	else
		self.deaths=self.deaths+1
		self.dc = self.orgdc
		self.status = 1
		self.so:stop()
		self.so:setAnimation(1)
	end
end

-------------------------------------------
-- ReggaeAmbassada is based on the same concept as GanjaWarrior but
-- leaves the map generation to functions in C, and uses an A* algorithm
-- coded in C too to choose the best path to the safe place
-------------------------------------------

ReggaeAmbassada = {
	so = nil, 			-- scriptobject
	status=STATUS_ALIVE, -- status
	orgspeed=250/1000,  -- travelling speed in px/s
	dc=0,				-- death counter
	orgdc=1000,			-- initial value for death counter
	lasttime=0,			-- time of last direction change
	deaths=0,			-- # times the Dumbo died
	maxdeaths=3, 		-- # times the Dumbo can die before its removed 
						-- from the field
	map=nil,			-- internal map
	newx=0,
	newy=0
}

function ReggaeAmbassada.new(script_obj)
	local self=createObj(ReggaeAmbassada)
	self.so=script_obj
	return self
end

function ReggaeAmbassada:here_is_safe()
	local bot=self.so
	local tx=Field:x_to_tile(bot:get_x())
	local ty=Field:y_to_tile(bot:get_y())
	return is_safe_place(tx,ty)
end

-- return coords of safe location from map
function ReggaeAmbassada:spiral_find()
	local tx=Field:x_to_tile(self.so:get_x())
	local ty=Field:y_to_tile(self.so:get_y())
	
	local dir=1  
	local nx=tx      -- tile we check
	local ny=ty-1
	local c=0       -- 
	local C=0 		--
	local n=0 		-- 
	local N=24 
	local S=0
	local dx,dy
	dx=0
	dy=0
	
	write('starting spiral from tx='..tx..' ty='..ty..'\n')
	
	while n<N do
		for s=0,S do
			for c=0,C do
				local v=self.map:get(nx,ny)
				if self.map:get(nx,ny) == 0 then
					write('found first safe heaven at nx='..nx..' ny='..ny..'\n')
					return nx,ny
				end
				dx,dy=dir2vect(dir)
				nx=nx+dx
				ny=ny+dy
				n=n+1
			end
			dir=mod(dir+1,4)
			if n>N then break end
		end
		C=C+1
		S=1
	end

	return nil,nil
end

-- find a route to a safe place
function ReggaeAmbassada:plan_escape()
	write('planning escape\n')
	
	local bot = self.so
	local tx = Field:x_to_tile(bot:get_x())
	local ty = Field:y_to_tile(bot:get_y())
	if self.map == nil then
		self.map = BotHelper:CreateProjection(10, 10, tx-5,ty-5)
	else
		self.map:set_fieldcoords(tx-5,ty-5)
	end
		
	write('created projection\n')
	self.map:dump()
	
	write('spiral to safe place\n')
	local gx,gy
	gx,gy=self:spiral_find()
	if gx==nil then
		print('spiral found shit.. random something up')
		self.starting=STATUS_IMPROVISE
		return
	end
	
	print('finding path from ',tx,ty,' to ',gx,gy)
	if not self.map:find_path(tx,ty,gx,gy) then
		write('pathfinder didn\'t find anything, resorting to random\n')
		BotHelper:FreeProjection(self.map)
		self.map = nil
		self.status = STATUS_ALIVE
		return
	end

--	write('-------------------------------\n')
--	write('would be path..\n')
--	while self.map:next_path_step() do
--		local x=self.map:get_nx()
--		local y=self.map:get_ny()
--		write(x..','..y..'\n')
--	end
--	write('-------------------------------\n')
--	self.map:rewind_path()
	
	-- skip 1st entry, should be current location
	self.map:next_path_step()

	-- set 1st target
	self.map:next_path_step()
	self.safe_x = self.map:get_nx()
	self.safe_y = self.map:get_ny()
	write('pathfinder returned '..self.safe_x..', '..self.safe_y..'\n')
			
	-- deduce direction and slide bomb
	local dx=tx - self.safe_x
	local dy=ty - self.safe_y

	if abs(dx)>abs(dy) then
		write('horizontal slide to '..self.safe_x..'\n')
		bot:horizontal_slide(self.orgspeed, self.safe_x*40+20)
	elseif abs(dx)<abs(dy) then
		write('vertical slide to '..self.safe_y..'\n')
		bot:vertical_slide(self.orgspeed, self.safe_y*40+20)
	end
end

function ReggaeAmbassada:think()
	local bot=self.so
	local tx=Field:x_to_tile(bot:get_x())
	local ty=Field:y_to_tile(bot:get_y())
	write('bot is at '..tx..','..ty..'\n')

	if self.map == nil then
		self.map = BotHelper:CreateProjection(10, 10, tx, ty);
--	else
--		self.map:set_fieldcoords(tx-5, ty-5)
--		self.map:update()
	end

--	self.map:dump()
	
	if self.status == STATUS_IMPROVISE then
		for i=0,1 do
			for i=0,3 do
				bot:slide(i,self.orgspeed,40)
				if bot:is_moving() then break end
			end
			if bot:is_moving() then break end
		end
		write('status was improvise, resetting to ALIVE\n')
		self.status = STATUS_ALIVE
		return
	end
	
	if self.status == STATUS_ALIVE then
		if self.lasttime == nil or Game:get_ticks() - self.lasttime > 000 then
			print('rechecking safety')
		
			if not self:here_is_safe() then
				write('uh oh.. here is not safe !\n')
				bot:stop()
				self:plan_escape()
				self.status = STATUS_FLEEING
			else
--				write('this place looks safe..\n')
			end
			self.lasttime = Game:get_ticks()
		end

		return
	end

	if self.status == STATUS_FLEEING then
		----------------------------------------------------
		--if Game:get_ticks() - self.lasttime > 1000 then
		--	write('reset target after 1 second\n')
		--	bot:stop()
		--	self:find_safe_place()
		--end
		---------------------------------------------------- 
		
		-- bot is stopped while fleeing
		if self.status == STATUS_FLEEING and not bot:is_moving() then
			if tx ~= self.safe_x or 
			   ty ~= self.safe_y 
			then
				-- stopped before reaching target
				print('stopped prematurely at ',tx,ty,'replanning now')
				self.map:set_fieldcoords(tx-5, ty-5)
				self.map:update()
				self:plan_escape()
				return
			end
			
			-- target reached
			
			-- get next step
			if self.map:next_path_step() == nil then
				-- reached last step
				print('bot reached target, status back to ALIVE')
				self.status = STATUS_ALIVE
				return
			end

			-- get new coords of new step
			self.safe_x = self.map:get_nx()
			self.safe_y = self.map:get_ny()
			print('proceeding to next target:',self.safe_x,self.safe_y)

			-- deduce direction and slide bomb accordingly
			local dx=tx - self.safe_x
			local dy=ty - self.safe_y

			if abs(dx)>=abs(dy) then
				write('horizontal slide to '..self.safe_x..'\n')
				bot:horizontal_slide(self.orgspeed, self.safe_x*40+20)
			else
				write('vertical slide to '..self.safe_y..'\n')
				bot:vertical_slide(self.orgspeed, self.safe_y*40+20)
			end

			-- did move operation succeed ?
			if not bot:is_moving() then
				-- bomb is not moving, panic
				write('plan blocked, improvising\n')
				self.status = STATUS_IMPROVISE
				return
			end
		end
	end

	if self.status == STATUS_DEAD then
		self.dc=self.dc-Game:get_elapsed_time()
		if self.dc < 0 then 
			if self.deaths>=self.maxdeaths then
				-- die for good
				self.so:set_remove_flag()
			else
				-- live again
				self.status = 0
				bot:setAnimation(0)
			end
		end
	end
end

function ReggaeAmbassada:die()
	if self.status == STATUS_DEAD then
		-- don't die when already dead
		return
	else
		self.deaths=self.deaths+1
		self.dc = self.orgdc
		self.status = 1
		self.so:stop()
		self.so:setAnimation(1)
	end
end
	
	
