-- global constructor
function createObj(template, init) 
	local new = init or {}
	for k,v in template do 
		if not new[k] then new[k]=v end
	end
	return new
end

Teleport = {
	so = nil,
	dest_x = nil,
	dest_y = nil
}

function addTeleport(x, y, _dest_x, _dest_y)
	local tp=createObj(Teleport, {dest_x=_dest_x*40+20, dest_y=_dest_y*40+20} )
	local tpref=lua_ref(tp)
	tp.so=Game:addScriptObject("Teleport", tpref, x*40+20, y*40+20)
	tp.so:setAnimation(2)
	return tp
end

function Teleport.new(script_obj)
	local self=createObj(Teleport)
	self.so=script_obj
	self.so:setAnimation(2)
	return self
end

function Teleport:think()
end

function Teleport:die()
end

-- teleport object to dest_x, dest_y or random_location
function Teleport:on_object_over(object)
	if self.dest_x ~= nil and 
	   self.dest_y ~= nil
	then
		object:set_x(self.dest_x)
		object:set_y(self.dest_y)
	else
		while 1==1 do
			local rx=random(1,Field:get_width()-1)
			local ry=random(1,Field:get_height()-1)
			if Field:can_cross(rx,ry) and
			   Field:get_object_at(rx,ry) == nil
			then
				object:set_x(Field:tile_to_x(rx))
				object:set_y(Field:tile_to_y(ry))
				break
			end
		end
	end
end

-- let players walk over this object and block other objects
function Teleport:blocks(object)
	if object:type() == OBJ_PLAYER then
		return nil
	else
		return 1
	end
end

-- Gate
-- consists of Gate, GateTop, and if necessary Switch
-- GateTop is a slave object created by Gate to handle the superior part of the gfx

Gate = {
	so = nil,
	is_open = 1,
	top = nil,
	status = 0,
	lasttime = nil,
	open_time = 500
}

GateTop = {
	so = nil,
	is_open = 1
}

function addGate(x, y)
	local tp=createObj(Gate)
	local tpref=lua_ref(tp)
	tp.so=Game:addScriptObject("Gate", tpref, x*40+20, y*40+20)
	tp.so:setAnimation(3)

	local tpt=createObj(GateTop)
	local tptref=lua_ref(tpt)
	tpt.so=Game:addScriptObject("GateTop", tptref, x*40+20, y*40-20)
	tpt.so:setAnimation(4)

	tp.top = tpt
	return tp
end

function Gate:think()
	if self.status == 0 then 
		-- status = idle
		return 
	end

	if self.status == 1 then 
		-- opening
		if Game:get_ticks() - self.lasttime > self.open_time then
			self.status = 0
			self.is_open = 1
			write('end open door\n')
		end
	elseif self.status == 2 then
		-- closing
		if Game:get_ticks() - self.lasttime > self.open_time then
			self.status = 0
			self.is_open = nil
			write('end close door\n')
		end
	end
end

function Gate:die()
end

function Gate:on_object_over(object)
	if self.status == 2 then
		-- kill objects in opening if status == closing
		object:die()
	end
end

function Gate:blocks(object)
	if self.is_open then 
		return nil
	else 
		return 1
	end
end

function Gate:open()
	if self.is_open or self.status ~= 0 then return end
	self.is_open = 1
	self.top.is_open = 1
	self.so:setAnimation(3)
	self.top.so:setAnimation(4)
	self.status=1
	self.lasttime = Game:get_ticks()
end

function Gate:close()
	if not self.is_open or self.status ~= 0 then return end
	self.is_open = nil
	self.top.is_open = nil
	self.so:setAnimation(5)
	self.top.so:setAnimation(6)
	self.status = 2
	self.lasttime = Game:get_ticks()
end

function Gate:toggle()
	if self.is_open then
		self:close()
	else
		self:open()
	end
end

function GateTop:think()
end

function GateTop:die()
end

function GateTop:on_object_over(go)
end

function GateTop:blocks(go)
	if self.is_open then return nil
	else return 1
	end
end

Switch = {
	so = nil,
	target = nil,
}

function addSwitch(x, y, _target)
	local tp=createObj(Teleport, {target=_target})
	local tpref=lua_ref(tp)
	tp.so=Game:addScriptObject("Switch", tpref, x*40+20, y*40+20)
	tp.so:setAnimation(7)
	return tp
end

function Switch:think()
	if self.target.is_open then
		self.so:setAnimation(7)
	else
		self.so:setAnimation(8)
	end

end

function Switch:die()
end

function Switch:on_object_over(object)
	if self.target then
		self.target:toggle()
	end
end

-- let everything pass
function Switch:blocks(object)
	return nil
end

